from django.test import TransactionTestCase
from django_dynamic_fixture import G
from entity.models import Entity, EntityRelationship

from entity_history.models import EntityActivationEvent, EntityRelationshipActivationEvent

from entity_history.sql.triggers import (
    EntityActivationTrigger,
    EntityRelationshipActivationTrigger,
    EntityRelationshipActivationImmediateTrigger
)


class EntityActivationTriggerTest(TransactionTestCase):
    def test_enable(self):
        # Enable the trigger
        EntityActivationTrigger().enable()

        # Create an entity
        G(Entity, is_active=True)

        # Assert that we have an entity activation event
        self.assertTrue(EntityActivationEvent.objects.count())

    def test_disable(self):
        # Enable the trigger
        EntityActivationTrigger().disable()

        # Create an entity
        G(Entity, is_active=True)

        # Assert that we have an entity activation event
        self.assertFalse(EntityActivationEvent.objects.count())


class EntityRelationshipActivationTriggerTest(TransactionTestCase):
    def test_enable(self):
        # Enable the trigger and assert an error is thrown
        with self.assertRaises(Exception):
            EntityRelationshipActivationTrigger().enable()

    def test_disable(self):
        # Enable the trigger
        EntityRelationshipActivationTrigger().disable()

        # Create an entity relationship
        G(EntityRelationship)

        # Assert that we do not have an entity activation event
        self.assertFalse(EntityRelationshipActivationEvent.objects.count())


class EntityRelationshipActivationImmediateTriggerTest(TransactionTestCase):
    def test_enable(self):
        # Enable the trigger
        EntityRelationshipActivationImmediateTrigger().enable()

        # Create an entity relationship
        G(EntityRelationship)

        # Assert that we have an entity activation event
        self.assertTrue(EntityRelationshipActivationEvent.objects.count())

    def test_disable(self):
        # Enable the trigger
        EntityRelationshipActivationImmediateTrigger().disable()

        # Create an entity relationship
        G(EntityRelationship)

        # Assert that we do not have an entity activation event
        self.assertFalse(EntityRelationshipActivationEvent.objects.count())



from datetime import datetime

from django.test import TestCase
from django_dynamic_fixture import G
from entity.models import Entity

from entity_history.models import (
    get_sub_entities_at_times, EntityRelationshipActivationEvent, get_entities_at_times, EntityActivationEvent,
    EntityHistory
)


class EntityManagerTest(TestCase):
    def test_all_entity_history_manager_returns_active_and_inactive(self):
        active_e = G(EntityHistory, is_active=True)
        inactive_e = G(EntityHistory, is_active=False)
        self.assertEquals(set([active_e, inactive_e]), set(EntityHistory.all_objects.all()))

    def test_active_entity_history_manager_returns_active(self):
        active_e = G(EntityHistory, is_active=True)
        G(Entity, is_active=False)
        self.assertEquals(set([active_e]), set(EntityHistory.objects.all()))


class GetSubEntitiesAtTimesTest(TestCase):
    """
    Test the get_sub_entities_at_times function.
    """
    def test_no_events_no_input(self):
        res = get_sub_entities_at_times([], [])
        self.assertEquals(res, {})

    def test_no_events_w_input(self):
        res = get_sub_entities_at_times([1, 2], [datetime(2013, 4, 5), datetime(2013, 5, 6)])
        self.assertEquals(res, {
            (1, datetime(2013, 4, 5)): set(),
            (1, datetime(2013, 5, 6)): set(),
            (2, datetime(2013, 4, 5)): set(),
            (2, datetime(2013, 5, 6)): set(),
        })

    def test_w_events_no_results(self):
        se = G(Entity)
        G(EntityRelationshipActivationEvent, was_activated=True, super_entity=se, time=datetime(2013, 2, 1))
        G(EntityRelationshipActivationEvent, was_activated=False, super_entity=se, time=datetime(2013, 2, 2))

        res = get_sub_entities_at_times([se.id], [datetime(2012, 4, 5), datetime(2012, 5, 6)])
        self.assertEquals(res, {
            (se.id, datetime(2012, 4, 5)): set(),
            (se.id, datetime(2012, 5, 6)): set(),
        })

    def test_w_events_one_sub_e_returned(self):
        super_e = G(Entity)
        sub_e = G(Entity)
        G(
            EntityRelationshipActivationEvent, was_activated=True, super_entity=super_e, sub_entity=sub_e,
            time=datetime(2013, 2, 1))
        G(
            EntityRelationshipActivationEvent, was_activated=False, super_entity=super_e, sub_entity=sub_e,
            time=datetime(2013, 2, 3))

        res = get_sub_entities_at_times([super_e.id], [datetime(2013, 2, 2), datetime(2012, 5, 6)])
        self.assertEquals(res, {
            (super_e.id, datetime(2013, 2, 2)): set([sub_e.id]),
            (super_e.id, datetime(2012, 5, 6)): set(),
        })

    def test_w_events_sub_entity_deactivated_before_date(self):
        super_e = G(Entity)
        sub_e = G(Entity)
        G(
            EntityRelationshipActivationEvent, was_activated=True, super_entity=super_e, sub_entity=sub_e,
            time=datetime(2013, 2, 1))
        G(
            EntityRelationshipActivationEvent, was_activated=False, super_entity=super_e, sub_entity=sub_e,
            time=datetime(2013, 2, 3))

        res = get_sub_entities_at_times([super_e.id], [datetime(2013, 2, 4)])
        self.assertEquals(res, {
            (super_e.id, datetime(2013, 2, 4)): set(),
        })

    def test_w_mulitple_activation_events_one_sub_e_returned(self):
        super_e = G(Entity)
        sub_e = G(Entity)
        G(
            EntityRelationshipActivationEvent, was_activated=True, super_entity=super_e, sub_entity=sub_e,
            time=datetime(2013, 2, 1))
        G(
            EntityRelationshipActivationEvent, was_activated=False, super_entity=super_e, sub_entity=sub_e,
            time=datetime(2013, 2, 3))
        G(
            EntityRelationshipActivationEvent, was_activated=False, super_entity=super_e, sub_entity=sub_e,
            time=datetime(2013, 2, 4))
        G(
            EntityRelationshipActivationEvent, was_activated=True, super_entity=super_e, sub_entity=sub_e,
            time=datetime(2013, 2, 4, 12))
        G(
            EntityRelationshipActivationEvent, was_activated=False, super_entity=super_e, sub_entity=sub_e,
            time=datetime(2013, 3, 4, 12))

        res = get_sub_entities_at_times([super_e.id], [datetime(2013, 2, 6), datetime(2012, 5, 6)])
        self.assertEquals(res, {
            (super_e.id, datetime(2013, 2, 6)): set([sub_e.id]),
            (super_e.id, datetime(2012, 5, 6)): set(),
        })

    def test_w_mulitple_activation_events_mulitple_sub_e_returned(self):
        super_e = G(Entity)
        sub_e1 = G(Entity)
        sub_e2 = G(Entity)
        G(
            EntityRelationshipActivationEvent, was_activated=True, super_entity=super_e, sub_entity=sub_e1,
            time=datetime(2013, 2, 1))
        G(
            EntityRelationshipActivationEvent, was_activated=False, super_entity=super_e, sub_entity=sub_e1,
            time=datetime(2013, 2, 3))
        G(
            EntityRelationshipActivationEvent, was_activated=False, super_entity=super_e, sub_entity=sub_e1,
            time=datetime(2013, 2, 4))
        G(
            EntityRelationshipActivationEvent, was_activated=True, super_entity=super_e, sub_entity=sub_e1,
            time=datetime(2013, 2, 4, 12))
        G(
            EntityRelationshipActivationEvent, was_activated=False, super_entity=super_e, sub_entity=sub_e1,
            time=datetime(2013, 3, 4, 12))

        G(
            EntityRelationshipActivationEvent, was_activated=True, super_entity=super_e, sub_entity=sub_e2,
            time=datetime(2013, 2, 4))
        G(
            EntityRelationshipActivationEvent, was_activated=False, super_entity=super_e, sub_entity=sub_e2,
            time=datetime(2013, 2, 20))
        G(
            EntityRelationshipActivationEvent, was_activated=False, super_entity=super_e, sub_entity=sub_e2,
            time=datetime(2013, 3, 4))
        G(
            EntityRelationshipActivationEvent, was_activated=True, super_entity=super_e, sub_entity=sub_e2,
            time=datetime(2013, 3, 4, 12))
        G(
            EntityRelationshipActivationEvent, was_activated=True, super_entity=super_e, sub_entity=sub_e2,
            time=datetime(2013, 3, 4, 13))

        res = get_sub_entities_at_times(
            [super_e.id], [datetime(2013, 2, 2), datetime(2013, 2, 4, 13), datetime(2013, 3, 5)])

        self.assertEquals(res, {
            (super_e.id, datetime(2013, 2, 2)): set([sub_e1.id]),
            (super_e.id, datetime(2013, 2, 4, 13)): set([sub_e1.id, sub_e2.id]),
            (super_e.id, datetime(2013, 3, 5)): set([sub_e2.id]),
        })

    def test_w_mulitple_activation_events_mulitple_sub_e_returned_w_filter(self):
        super_e = G(Entity)
        sub_e1 = G(Entity)
        sub_e2 = G(Entity)
        G(
            EntityRelationshipActivationEvent, was_activated=True, super_entity=super_e, sub_entity=sub_e1,
            time=datetime(2013, 2, 1))
        G(
            EntityRelationshipActivationEvent, was_activated=False, super_entity=super_e, sub_entity=sub_e1,
            time=datetime(2013, 2, 3))
        G(
            EntityRelationshipActivationEvent, was_activated=False, super_entity=super_e, sub_entity=sub_e1,
            time=datetime(2013, 2, 4))
        G(
            EntityRelationshipActivationEvent, was_activated=True, super_entity=super_e, sub_entity=sub_e1,
            time=datetime(2013, 2, 4, 12))
        G(
            EntityRelationshipActivationEvent, was_activated=False, super_entity=super_e, sub_entity=sub_e1,
            time=datetime(2013, 3, 4, 12))

        G(
            EntityRelationshipActivationEvent, was_activated=True, super_entity=super_e, sub_entity=sub_e2,
            time=datetime(2013, 2, 4))
        G(
            EntityRelationshipActivationEvent, was_activated=False, super_entity=super_e, sub_entity=sub_e2,
            time=datetime(2013, 2, 20))
        G(
            EntityRelationshipActivationEvent, was_activated=False, super_entity=super_e, sub_entity=sub_e2,
            time=datetime(2013, 3, 4))
        G(
            EntityRelationshipActivationEvent, was_activated=True, super_entity=super_e, sub_entity=sub_e2,
            time=datetime(2013, 3, 4, 12))
        G(
            EntityRelationshipActivationEvent, was_activated=True, super_entity=super_e, sub_entity=sub_e2,
            time=datetime(2013, 3, 4, 13))

        res = get_sub_entities_at_times(
            [super_e.id], [datetime(2013, 2, 2), datetime(2013, 2, 4, 13), datetime(2013, 3, 5)],
            filter_by_entity_ids=[sub_e2.id])

        self.assertEquals(res, {
            (super_e.id, datetime(2013, 2, 2)): set(),
            (super_e.id, datetime(2013, 2, 4, 13)): set([sub_e2.id]),
            (super_e.id, datetime(2013, 3, 5)): set([sub_e2.id]),
        })

    def test_w_mulitple_activation_events_mulitple_sub_e_returned_w_queryset_filter(self):
        super_e = G(Entity)
        sub_e1 = G(Entity)
        sub_e2 = G(Entity)
        G(
            EntityRelationshipActivationEvent, was_activated=True, super_entity=super_e, sub_entity=sub_e1,
            time=datetime(2013, 2, 1))
        G(
            EntityRelationshipActivationEvent, was_activated=False, super_entity=super_e, sub_entity=sub_e1,
            time=datetime(2013, 2, 3))
        G(
            EntityRelationshipActivationEvent, was_activated=False, super_entity=super_e, sub_entity=sub_e1,
            time=datetime(2013, 2, 4))
        G(
            EntityRelationshipActivationEvent, was_activated=True, super_entity=super_e, sub_entity=sub_e1,
            time=datetime(2013, 2, 4, 12))
        G(
            EntityRelationshipActivationEvent, was_activated=False, super_entity=super_e, sub_entity=sub_e1,
            time=datetime(2013, 3, 4, 12))

        G(
            EntityRelationshipActivationEvent, was_activated=True, super_entity=super_e, sub_entity=sub_e2,
            time=datetime(2013, 2, 4))
        G(
            EntityRelationshipActivationEvent, was_activated=False, super_entity=super_e, sub_entity=sub_e2,
            time=datetime(2013, 2, 20))
        G(
            EntityRelationshipActivationEvent, was_activated=False, super_entity=super_e, sub_entity=sub_e2,
            time=datetime(2013, 3, 4))
        G(
            EntityRelationshipActivationEvent, was_activated=True, super_entity=super_e, sub_entity=sub_e2,
            time=datetime(2013, 3, 4, 12))
        G(
            EntityRelationshipActivationEvent, was_activated=True, super_entity=super_e, sub_entity=sub_e2,
            time=datetime(2013, 3, 4, 13))

        res = EntityHistory.objects.filter(id=sub_e2.id).get_sub_entities_at_times(
            [super_e.id], [datetime(2013, 2, 2), datetime(2013, 2, 4, 13), datetime(2013, 3, 5)])

        self.assertEquals(res, {
            (super_e.id, datetime(2013, 2, 2)): set(),
            (super_e.id, datetime(2013, 2, 4, 13)): set([sub_e2.id]),
            (super_e.id, datetime(2013, 3, 5)): set([sub_e2.id]),
        })

    def test_w_mulitple_activation_events_mulitple_sub_e_returned_w_manager(self):
        super_e = G(Entity)
        sub_e1 = G(Entity)
        sub_e2 = G(Entity)
        G(
            EntityRelationshipActivationEvent, was_activated=True, super_entity=super_e, sub_entity=sub_e1,
            time=datetime(2013, 2, 1))
        G(
            EntityRelationshipActivationEvent, was_activated=False, super_entity=super_e, sub_entity=sub_e1,
            time=datetime(2013, 2, 3))
        G(
            EntityRelationshipActivationEvent, was_activated=False, super_entity=super_e, sub_entity=sub_e1,
            time=datetime(2013, 2, 4))
        G(
            EntityRelationshipActivationEvent, was_activated=True, super_entity=super_e, sub_entity=sub_e1,
            time=datetime(2013, 2, 4, 12))
        G(
            EntityRelationshipActivationEvent, was_activated=False, super_entity=super_e, sub_entity=sub_e1,
            time=datetime(2013, 3, 4, 12))

        G(
            EntityRelationshipActivationEvent, was_activated=True, super_entity=super_e, sub_entity=sub_e2,
            time=datetime(2013, 2, 4))
        G(
            EntityRelationshipActivationEvent, was_activated=False, super_entity=super_e, sub_entity=sub_e2,
            time=datetime(2013, 2, 20))
        G(
            EntityRelationshipActivationEvent, was_activated=False, super_entity=super_e, sub_entity=sub_e2,
            time=datetime(2013, 3, 4))
        G(
            EntityRelationshipActivationEvent, was_activated=True, super_entity=super_e, sub_entity=sub_e2,
            time=datetime(2013, 3, 4, 12))
        G(
            EntityRelationshipActivationEvent, was_activated=True, super_entity=super_e, sub_entity=sub_e2,
            time=datetime(2013, 3, 4, 13))

        res = EntityHistory.objects.get_sub_entities_at_times(
            [super_e.id], [datetime(2013, 2, 2), datetime(2013, 2, 4, 13), datetime(2013, 3, 5)])

        self.assertEquals(res, {
            (super_e.id, datetime(2013, 2, 2)): set([sub_e1.id]),
            (super_e.id, datetime(2013, 2, 4, 13)): set([sub_e1.id, sub_e2.id]),
            (super_e.id, datetime(2013, 3, 5)): set([sub_e2.id]),
        })

    def test_w_mulitple_activation_events_mulitple_super_e_and_sub_e_returned(self):
        super_e1 = G(Entity)
        super_e2 = G(Entity)
        sub_e1 = G(Entity)
        sub_e2 = G(Entity)
        G(
            EntityRelationshipActivationEvent, was_activated=True, super_entity=super_e1, sub_entity=sub_e1,
            time=datetime(2013, 2, 1))
        G(
            EntityRelationshipActivationEvent, was_activated=False, super_entity=super_e1, sub_entity=sub_e1,
            time=datetime(2013, 2, 3))
        G(
            EntityRelationshipActivationEvent, was_activated=False, super_entity=super_e1, sub_entity=sub_e1,
            time=datetime(2013, 2, 4))
        G(
            EntityRelationshipActivationEvent, was_activated=True, super_entity=super_e1, sub_entity=sub_e1,
            time=datetime(2013, 2, 4, 12))
        G(
            EntityRelationshipActivationEvent, was_activated=False, super_entity=super_e1, sub_entity=sub_e1,
            time=datetime(2013, 3, 4, 12))
        G(
            EntityRelationshipActivationEvent, was_activated=True, super_entity=super_e2, sub_entity=sub_e1,
            time=datetime(2013, 1, 1))
        G(
            EntityRelationshipActivationEvent, was_activated=False, super_entity=super_e2, sub_entity=sub_e1,
            time=datetime(2013, 12, 1))

        G(
            EntityRelationshipActivationEvent, was_activated=True, super_entity=super_e1, sub_entity=sub_e2,
            time=datetime(2013, 2, 4))
        G(
            EntityRelationshipActivationEvent, was_activated=False, super_entity=super_e1, sub_entity=sub_e2,
            time=datetime(2013, 2, 20))
        G(
            EntityRelationshipActivationEvent, was_activated=False, super_entity=super_e1, sub_entity=sub_e2,
            time=datetime(2013, 3, 4))
        G(
            EntityRelationshipActivationEvent, was_activated=True, super_entity=super_e1, sub_entity=sub_e2,
            time=datetime(2013, 3, 4, 12))
        G(
            EntityRelationshipActivationEvent, was_activated=True, super_entity=super_e1, sub_entity=sub_e2,
            time=datetime(2013, 3, 4, 13))
        G(
            EntityRelationshipActivationEvent, was_activated=True, super_entity=super_e2, sub_entity=sub_e2,
            time=datetime(2013, 1, 1))
        G(
            EntityRelationshipActivationEvent, was_activated=False, super_entity=super_e2, sub_entity=sub_e2,
            time=datetime(2013, 12, 1))

        res = get_sub_entities_at_times(
            [super_e1.id, super_e2.id], [datetime(2013, 2, 2), datetime(2013, 2, 4, 13), datetime(2013, 3, 5)])

        self.assertEquals(res, {
            (super_e1.id, datetime(2013, 2, 2)): set([sub_e1.id]),
            (super_e1.id, datetime(2013, 2, 4, 13)): set([sub_e1.id, sub_e2.id]),
            (super_e1.id, datetime(2013, 3, 5)): set([sub_e2.id]),
            (super_e2.id, datetime(2013, 2, 2)): set([sub_e1.id, sub_e2.id]),
            (super_e2.id, datetime(2013, 2, 4, 13)): set([sub_e1.id, sub_e2.id]),
            (super_e2.id, datetime(2013, 3, 5)): set([sub_e1.id, sub_e2.id]),
        })


class GetEntitiesAtTimeTest(TestCase):
    """
    Test the get_entities_at_times function.
    """
    def test_no_events_no_input(self):
        res = get_entities_at_times([])
        self.assertEquals(res, {})

    def test_no_events_w_input(self):
        res = get_entities_at_times([datetime(2013, 4, 5), datetime(2013, 5, 6)])
        self.assertEquals(res, {
            datetime(2013, 4, 5): set(),
            datetime(2013, 5, 6): set(),
            datetime(2013, 4, 5): set(),
            datetime(2013, 5, 6): set(),
        })

    def test_w_events_no_results(self):
        e = G(Entity)
        G(EntityActivationEvent, was_activated=True, entity=e, time=datetime(2013, 2, 1))
        G(EntityActivationEvent, was_activated=False, entity=e, time=datetime(2013, 2, 2))

        res = get_entities_at_times([datetime(2012, 4, 5), datetime(2012, 5, 6)])
        self.assertEquals(res, {
            datetime(2012, 4, 5): set(),
            datetime(2012, 5, 6): set(),
        })

    def test_w_events_one_e_returned(self):
        e = G(Entity)
        G(EntityActivationEvent, was_activated=True, entity=e, time=datetime(2013, 2, 1))
        G(EntityActivationEvent, was_activated=False, entity=e, time=datetime(2013, 2, 3))

        res = get_entities_at_times([datetime(2013, 2, 2), datetime(2012, 5, 6)])
        self.assertEquals(res, {
            datetime(2013, 2, 2): set([e.id]),
            datetime(2012, 5, 6): set(),
        })

    def test_w_events_entity_deactivated_before_date(self):
        e = G(Entity)
        G(EntityActivationEvent, was_activated=True, entity=e, time=datetime(2013, 2, 1))
        G(EntityActivationEvent, was_activated=False, entity=e, time=datetime(2013, 2, 3))

        res = get_entities_at_times([datetime(2013, 2, 4)])
        self.assertEquals(res, {
            datetime(2013, 2, 4): set(),
        })

    def test_w_mulitple_activation_events_one_e_returned(self):
        e = G(Entity)
        G(EntityActivationEvent, was_activated=True, entity=e, time=datetime(2013, 2, 1))
        G(EntityActivationEvent, was_activated=False, entity=e, time=datetime(2013, 2, 3))
        G(EntityActivationEvent, was_activated=False, entity=e, time=datetime(2013, 2, 4))
        G(EntityActivationEvent, was_activated=True, entity=e, time=datetime(2013, 2, 4, 12))
        G(EntityActivationEvent, was_activated=False, entity=e, time=datetime(2013, 3, 4, 12))

        res = get_entities_at_times([datetime(2013, 2, 6), datetime(2012, 5, 6)])
        self.assertEquals(res, {
            datetime(2013, 2, 6): set([e.id]),
            datetime(2012, 5, 6): set(),
        })

    def test_w_mulitple_activation_events_mulitple_e_returned(self):
        e1 = G(Entity)
        e2 = G(Entity)
        G(EntityActivationEvent, was_activated=True, entity=e1, time=datetime(2013, 2, 1))
        G(EntityActivationEvent, was_activated=False, entity=e1, time=datetime(2013, 2, 3))
        G(EntityActivationEvent, was_activated=False, entity=e1, time=datetime(2013, 2, 4))
        G(EntityActivationEvent, was_activated=True, entity=e1, time=datetime(2013, 2, 4, 12))
        G(EntityActivationEvent, was_activated=False, entity=e1, time=datetime(2013, 3, 4, 12))

        G(EntityActivationEvent, was_activated=True, entity=e2, time=datetime(2013, 2, 4))
        G(EntityActivationEvent, was_activated=False, entity=e2, time=datetime(2013, 2, 20))
        G(EntityActivationEvent, was_activated=False, entity=e2, time=datetime(2013, 3, 4))
        G(EntityActivationEvent, was_activated=True, entity=e2, time=datetime(2013, 3, 4, 12))
        G(EntityActivationEvent, was_activated=True, entity=e2, time=datetime(2013, 3, 4, 13))

        res = get_entities_at_times([datetime(2013, 2, 2), datetime(2013, 2, 4, 13), datetime(2013, 3, 5)])

        self.assertEquals(res, {
            datetime(2013, 2, 2): set([e1.id]),
            datetime(2013, 2, 4, 13): set([e1.id, e2.id]),
            datetime(2013, 3, 5): set([e2.id]),
        })

    def test_w_mulitple_activation_events_mulitple_e_returned_w_filter(self):
        e1 = G(Entity)
        e2 = G(Entity)
        G(EntityActivationEvent, was_activated=True, entity=e1, time=datetime(2013, 2, 1))
        G(EntityActivationEvent, was_activated=False, entity=e1, time=datetime(2013, 2, 3))
        G(EntityActivationEvent, was_activated=False, entity=e1, time=datetime(2013, 2, 4))
        G(EntityActivationEvent, was_activated=True, entity=e1, time=datetime(2013, 2, 4, 12))
        G(EntityActivationEvent, was_activated=False, entity=e1, time=datetime(2013, 3, 4, 12))

        G(EntityActivationEvent, was_activated=True, entity=e2, time=datetime(2013, 2, 4))
        G(EntityActivationEvent, was_activated=False, entity=e2, time=datetime(2013, 2, 20))
        G(EntityActivationEvent, was_activated=False, entity=e2, time=datetime(2013, 3, 4))
        G(EntityActivationEvent, was_activated=True, entity=e2, time=datetime(2013, 3, 4, 12))
        G(EntityActivationEvent, was_activated=True, entity=e2, time=datetime(2013, 3, 4, 13))

        res = get_entities_at_times(
            [datetime(2013, 2, 2), datetime(2013, 2, 4, 13), datetime(2013, 3, 5)],
            filter_by_entity_ids=[e1.id])

        self.assertEquals(res, {
            datetime(2013, 2, 2): set([e1.id]),
            datetime(2013, 2, 4, 13): set([e1.id]),
            datetime(2013, 3, 5): set(),
        })

    def test_w_mulitple_activation_events_mulitple_e_returned_w_queryset_filter(self):
        e1 = G(Entity)
        e2 = G(Entity)
        G(EntityActivationEvent, was_activated=True, entity=e1, time=datetime(2013, 2, 1))
        G(EntityActivationEvent, was_activated=False, entity=e1, time=datetime(2013, 2, 3))
        G(EntityActivationEvent, was_activated=False, entity=e1, time=datetime(2013, 2, 4))
        G(EntityActivationEvent, was_activated=True, entity=e1, time=datetime(2013, 2, 4, 12))
        G(EntityActivationEvent, was_activated=False, entity=e1, time=datetime(2013, 3, 4, 12))

        G(EntityActivationEvent, was_activated=True, entity=e2, time=datetime(2013, 2, 4))
        G(EntityActivationEvent, was_activated=False, entity=e2, time=datetime(2013, 2, 20))
        G(EntityActivationEvent, was_activated=False, entity=e2, time=datetime(2013, 3, 4))
        G(EntityActivationEvent, was_activated=True, entity=e2, time=datetime(2013, 3, 4, 12))
        G(EntityActivationEvent, was_activated=True, entity=e2, time=datetime(2013, 3, 4, 13))

        res = EntityHistory.objects.filter(id=e1.id).get_entities_at_times(
            [datetime(2013, 2, 2), datetime(2013, 2, 4, 13), datetime(2013, 3, 5)])

        self.assertEquals(res, {
            datetime(2013, 2, 2): set([e1.id]),
            datetime(2013, 2, 4, 13): set([e1.id]),
            datetime(2013, 3, 5): set(),
        })

    def test_w_mulitple_activation_events_mulitple_e_returned_w_manager(self):
        e1 = G(Entity)
        e2 = G(Entity)
        G(EntityActivationEvent, was_activated=True, entity=e1, time=datetime(2013, 2, 1))
        G(EntityActivationEvent, was_activated=False, entity=e1, time=datetime(2013, 2, 3))
        G(EntityActivationEvent, was_activated=False, entity=e1, time=datetime(2013, 2, 4))
        G(EntityActivationEvent, was_activated=True, entity=e1, time=datetime(2013, 2, 4, 12))
        G(EntityActivationEvent, was_activated=False, entity=e1, time=datetime(2013, 3, 4, 12))

        G(EntityActivationEvent, was_activated=True, entity=e2, time=datetime(2013, 2, 4))
        G(EntityActivationEvent, was_activated=False, entity=e2, time=datetime(2013, 2, 20))
        G(EntityActivationEvent, was_activated=False, entity=e2, time=datetime(2013, 3, 4))
        G(EntityActivationEvent, was_activated=True, entity=e2, time=datetime(2013, 3, 4, 12))
        G(EntityActivationEvent, was_activated=True, entity=e2, time=datetime(2013, 3, 4, 13))

        res = EntityHistory.objects.get_entities_at_times(
            [datetime(2013, 2, 2), datetime(2013, 2, 4, 13), datetime(2013, 3, 5)])

        self.assertEquals(res, {
            datetime(2013, 2, 2): set([e1.id]),
            datetime(2013, 2, 4, 13): set([e1.id, e2.id]),
            datetime(2013, 3, 5): set([e2.id]),
        })



from datetime import datetime

from django.test import TestCase
from django_dynamic_fixture import G, N
from entity.models import Entity

from entity_history.models import EntityActivationEvent


class EntityActivationTriggerTests(TestCase):
    """
    Tests that entity activation events are properly created when entities
    are activated and deactivated. This is accomplished by a postgres database
    trigger that is installed in a data migration.
    """
    def test_bulk_create(self):
        # Create the entities
        es = [N(Entity, is_active=True) for i in range(3)]
        t1 = datetime.utcnow()
        Entity.objects.bulk_create(es)
        t2 = datetime.utcnow()
        es = list(Entity.objects.order_by('id'))

        # Check that the events exist
        events = list(EntityActivationEvent.objects.all())
        self.assertTrue(len(events), 3)
        for i in range(3):
            self.assertEquals(events[i].entity, es[i])
            self.assertTrue(t1 <= events[i].time <= t2)

    def test_bulk_create_update(self):
        es = [N(Entity, is_active=True) for i in range(3)]
        t1 = datetime.utcnow()
        Entity.objects.bulk_create(es)
        t2 = datetime.utcnow()
        es = list(Entity.objects.order_by('id'))

        Entity.objects.update(is_active=False)
        t3 = datetime.utcnow()

        events = list(EntityActivationEvent.objects.all())
        self.assertTrue(len(events), 6)
        for i in range(3):
            self.assertEquals(events[i].entity, es[i])
            self.assertTrue(t1 <= events[i].time <= t2)
            self.assertTrue(events[i].was_activated)

        # The entities deactivation order is dependent on the DB. Populate
        # all deactivated entities and verify they are the same as the entities
        deactivated = []
        for i in range(3, 6):
            self.assertTrue(t2 <= events[i].time <= t3)
            self.assertFalse(events[i].was_activated)
            deactivated.append(events[i].entity)
        self.assertEquals(set(deactivated), set(es))

    def test_entity_creation_activated(self):
        t1 = datetime.utcnow()
        e = G(Entity, is_active=True)
        t2 = datetime.utcnow()

        event = EntityActivationEvent.objects.get()
        self.assertTrue(event.was_activated)
        self.assertEquals(event.entity, e)
        self.assertTrue(t1 <= event.time <= t2)

    def test_entity_creation_deactivated(self):
        t1 = datetime.utcnow()
        e = G(Entity, is_active=False)
        t2 = datetime.utcnow()

        event = EntityActivationEvent.objects.get()
        self.assertFalse(event.was_activated)
        self.assertEquals(event.entity, e)
        self.assertTrue(t1 <= event.time <= t2)

    def test_update_inactive_to_active(self):
        t1 = datetime.utcnow()
        e = G(Entity, is_active=False)
        t2 = datetime.utcnow()
        e.is_active = True
        e.save()
        t3 = datetime.utcnow()

        events = list(EntityActivationEvent.objects.order_by('time', 'id'))

        self.assertFalse(events[0].was_activated)
        self.assertEquals(events[0].entity, e)
        self.assertTrue(t1 <= events[0].time <= t2)

        self.assertTrue(events[1].was_activated)
        self.assertEquals(events[1].entity, e)
        self.assertTrue(t2 <= events[1].time <= t3)

    def test_update_inactive_to_active_to_active(self):
        t1 = datetime.utcnow()
        e = G(Entity, is_active=False)
        t2 = datetime.utcnow()
        e.is_active = True
        e.save()
        t3 = datetime.utcnow()
        e.is_active = True
        e.save()

        events = list(EntityActivationEvent.objects.order_by('time', 'id'))

        self.assertFalse(events[0].was_activated)
        self.assertEquals(events[0].entity, e)
        self.assertTrue(t1 <= events[0].time <= t2)

        self.assertTrue(events[1].was_activated)
        self.assertEquals(events[1].entity, e)
        self.assertTrue(t2 <= events[1].time <= t3)

    def test_update_inactive_to_active_to_active_to_inactive(self):
        t1 = datetime.utcnow()
        e = G(Entity, is_active=False)
        t2 = datetime.utcnow()
        e.is_active = True
        e.save()
        t3 = datetime.utcnow()
        e.is_active = True
        e.save()
        e.is_active = False
        e.save()
        t4 = datetime.utcnow()

        events = list(EntityActivationEvent.objects.order_by('time', 'id'))

        self.assertFalse(events[0].was_activated)
        self.assertEquals(events[0].entity, e)
        self.assertTrue(t1 <= events[0].time <= t2)

        self.assertTrue(events[1].was_activated)
        self.assertEquals(events[1].entity, e)
        self.assertTrue(t2 <= events[1].time <= t3)

        self.assertFalse(events[2].was_activated)
        self.assertEquals(events[2].entity, e)
        self.assertTrue(t3 <= events[2].time <= t4)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.9 on 2018-01-08 20:21

import datetime
from django.db import migrations, models
from entity_history.sql.foreign_keys import EntityHistoryForeignKeys

from entity_history.sql.triggers import EntityActivationTrigger, EntityRelationshipActivationImmediateTrigger


def refresh_entity_activation_trigger(*args, **kwargs):
    EntityActivationTrigger().disable()
    EntityActivationTrigger().enable()


def refresh_entity_relationship_activation_trigger(*args, **kwargs):
    EntityRelationshipActivationImmediateTrigger().disable()
    EntityRelationshipActivationImmediateTrigger().enable()


def update_foreign_keys(*args, **kwargs):
    EntityHistoryForeignKeys().enable_cascade_delete()


class Migration(migrations.Migration):

    dependencies = [
        ('entity_history', '0004_update_triggers_and_foreign_keys'),
    ]

    operations = [
        migrations.AddField(
            model_name='entityactivationevent',
            name='creation_time',
            field=models.DateTimeField(db_index=True, default=datetime.datetime.utcnow, help_text=b'The time that this entry was created'),
        ),
        migrations.AddField(
            model_name='entityactivationevent',
            name='updated_time',
            field=models.DateTimeField(db_index=True, default=datetime.datetime.utcnow, help_text=b'The time that this entry was updated'),
        ),
        migrations.AddField(
            model_name='entityrelationshipactivationevent',
            name='creation_time',
            field=models.DateTimeField(db_index=True, default=datetime.datetime.utcnow, help_text=b'The time that this entry was created'),
        ),
        migrations.AddField(
            model_name='entityrelationshipactivationevent',
            name='updated_time',
            field=models.DateTimeField(db_index=True, default=datetime.datetime.utcnow, help_text=b'The time that this entry was updated'),
        ),
        migrations.AlterField(
            model_name='entityactivationevent',
            name='time',
            field=models.DateTimeField(db_index=True, default=datetime.datetime.utcnow, help_text=b'The time of the activation / deactivation'),
        ),
        migrations.AlterField(
            model_name='entityrelationshipactivationevent',
            name='time',
            field=models.DateTimeField(db_index=True, default=datetime.datetime.utcnow, help_text=b'The time of the activation / deactivation'),
        ),
        migrations.RunPython(
            code=refresh_entity_activation_trigger,
        ),
        migrations.RunPython(
            code=refresh_entity_relationship_activation_trigger
        ),
        migrations.RunPython(
            code=update_foreign_keys
        )
    ]



# -*- coding: utf-8 -*-

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('entity_history', '0002_auto_20150406_1605'),
    ]

    operations = []



# -*- coding: utf-8 -*-

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('entity_history', '0001_initial'),
    ]

    operations = []



# -*- coding: utf-8 -*-

from django.db import models, migrations

from entity_history.sql.foreign_keys import EntityHistoryForeignKeys
from entity_history.sql.triggers import (
    EntityActivationTrigger,
    EntityRelationshipActivationImmediateTrigger
)


def refresh_entity_activation_trigger(*args, **kwargs):
    EntityActivationTrigger().disable()
    EntityActivationTrigger().enable()


def refresh_entity_relationship_activation_trigger(*args, **kwargs):
    EntityRelationshipActivationImmediateTrigger().disable()
    EntityRelationshipActivationImmediateTrigger().enable()


def update_foreign_keys(*args, **kwargs):
    EntityHistoryForeignKeys().enable_cascade_delete()


class Migration(migrations.Migration):

    dependencies = [
        ('entity_history', '0003_update_triggers'),
    ]

    operations = [
        migrations.RunPython(
            code=refresh_entity_activation_trigger,
        ),
        migrations.RunPython(
            code=refresh_entity_relationship_activation_trigger
        ),
        migrations.RunPython(
            code=update_foreign_keys
        )
    ]



from django.db import connection
from os.path import dirname


class SqlTrigger(object):
    trigger_procedure_create_name = None
    trigger_procedure_delete_name = None
    trigger_create_name = None
    trigger_delete_name = None

    def get_sql(self, name):
        return open(
            dirname(__file__) + '/' + name
        ).read()

    def enable(self):
        with connection.cursor() as cursor:
            cursor.execute(self.get_sql(self.trigger_procedure_create_name))
            cursor.execute(self.get_sql(self.trigger_create_name))

    def disable(self):
        with connection.cursor() as cursor:
            cursor.execute(self.get_sql(self.trigger_delete_name))
            cursor.execute(self.get_sql(self.trigger_procedure_delete_name))


class EntityActivationTrigger(SqlTrigger):
    trigger_procedure_create_name = 'entity_activation_procedure_create.sql'
    trigger_procedure_delete_name = 'entity_activation_procedure_delete.sql'
    trigger_create_name = 'entity_activation_trigger_create.sql'
    trigger_delete_name = 'entity_activation_trigger_delete.sql'


class EntityRelationshipActivationTrigger(SqlTrigger):
    trigger_procedure_create_name = 'entity_relationship_activation_procedure_create.sql'
    trigger_procedure_delete_name = 'entity_relationship_activation_procedure_delete.sql'
    trigger_create_name = 'entity_relationship_activation_trigger_create.sql'
    trigger_delete_name = 'entity_relationship_activation_trigger_delete.sql'

    def enable(self):
        raise Exception('This trigger is deprecated, use the immediate version')


class EntityRelationshipActivationImmediateTrigger(SqlTrigger):
    """
    This is an immediate version of the relationship activation trigger.
    """
    trigger_procedure_create_name = 'entity_relationship_activation_procedure_create.sql'
    trigger_procedure_delete_name = 'entity_relationship_activation_procedure_delete.sql'
    trigger_create_name = 'entity_relationship_activation_trigger_immediate_create.sql'
    trigger_delete_name = 'entity_relationship_activation_trigger_delete.sql'



from datetime import datetime

from django.test import TransactionTestCase
from django_dynamic_fixture import G, N
from entity.models import EntityRelationship, Entity

from entity_history.models import EntityRelationshipActivationEvent


class EntityRelationshipActivationTriggerTests(TransactionTestCase):
    """
    Tests that entity activation events are properly created when entities
    are activated and deactivated. This is accomplished by a postgres database
    trigger that is installed in a data migration.
    """
    def test_relationship_cascade_delete(self):
        """
        Tests the case when an entity is deleted, which causes the relationship to
        be cascaded. This should result in no history being stored since the entity
        is no longer present.
        """
        e = G(Entity)
        G(EntityRelationship, super_entity=e)
        e.delete(force=True)

        self.assertFalse(Entity.objects.filter(id=e.id).exists())
        self.assertFalse(EntityRelationshipActivationEvent.objects.exists())

    def test_bulk_create(self):
        ers = [N(EntityRelationship) for i in range(3)]
        t1 = datetime.utcnow()
        EntityRelationship.objects.bulk_create(ers)
        t2 = datetime.utcnow()

        events = list(EntityRelationshipActivationEvent.objects.all())
        self.assertEquals(len(events), 3)
        for i in range(3):
            self.assertTrue(events[i].was_activated)
            self.assertTrue(t1 <= events[i].time <= t2)

    def test_bulk_create_delete(self):
        ers = [N(EntityRelationship) for i in range(3)]
        t1 = datetime.utcnow()
        EntityRelationship.objects.bulk_create(ers)
        t2 = datetime.utcnow()
        EntityRelationship.objects.all().delete()
        t3 = datetime.utcnow()

        events = list(EntityRelationshipActivationEvent.objects.all())
        self.assertTrue(len(events), 3)
        for i in range(3):
            self.assertTrue(events[i].was_activated)
            self.assertTrue(t1 <= events[i].time <= t2)
        for i in range(3, 6):
            self.assertFalse(events[i].was_activated)
            self.assertTrue(t2 <= events[i].time <= t3)

    def test_entity_relationship_creation(self):
        t1 = datetime.utcnow()
        er = G(EntityRelationship)
        t2 = datetime.utcnow()

        event = EntityRelationshipActivationEvent.objects.get()
        self.assertTrue(event.was_activated)
        self.assertEquals(event.sub_entity, er.sub_entity)
        self.assertEquals(event.super_entity, er.super_entity)
        self.assertTrue(t1 <= event.time <= t2)

    def test_entity_relationship_creation_deletion(self):
        t1 = datetime.utcnow()
        er = G(EntityRelationship)
        sub_entity = er.sub_entity
        super_entity = er.super_entity
        t2 = datetime.utcnow()
        er.delete()
        t3 = datetime.utcnow()

        events = list(EntityRelationshipActivationEvent.objects.order_by('time', 'id'))

        self.assertTrue(events[0].was_activated)
        self.assertEquals(events[0].sub_entity, sub_entity)
        self.assertEquals(events[0].super_entity, super_entity)
        self.assertTrue(t1 <= events[0].time <= t2)

        self.assertFalse(events[1].was_activated)
        self.assertEquals(events[1].sub_entity, sub_entity)
        self.assertEquals(events[1].super_entity, super_entity)
        self.assertTrue(t2 <= events[1].time <= t3)

    def test_entity_relationship_creation_deletion_creation_save(self):
        t1 = datetime.utcnow()
        er = G(EntityRelationship)
        sub_entity1 = er.sub_entity
        super_entity1 = er.super_entity
        t2 = datetime.utcnow()
        er.delete()
        t3 = datetime.utcnow()
        er = G(EntityRelationship)
        sub_entity2 = er.sub_entity
        super_entity2 = er.super_entity
        er.save()
        t4 = datetime.utcnow()

        events = list(EntityRelationshipActivationEvent.objects.order_by('time', 'id'))

        self.assertTrue(events[0].was_activated)
        self.assertEquals(events[0].sub_entity, sub_entity1)
        self.assertEquals(events[0].super_entity, super_entity1)
        self.assertTrue(t1 <= events[0].time <= t2)

        self.assertFalse(events[1].was_activated)
        self.assertEquals(events[1].sub_entity, sub_entity1)
        self.assertEquals(events[1].super_entity, super_entity1)
        self.assertTrue(t2 <= events[1].time <= t3)

        self.assertTrue(events[2].was_activated)
        self.assertEquals(events[2].sub_entity, sub_entity2)
        self.assertEquals(events[2].super_entity, super_entity2)
        self.assertTrue(t3 <= events[2].time <= t4)

    def test_entity_relationship_multiple_joins(self):
        """
        Test the case when an entity joins and leaves a relationship multiple times
        """

        # Create the relation
        entity_relation = G(EntityRelationship)
        sub_entity = entity_relation.sub_entity
        super_entity = entity_relation.super_entity

        # Delete the relationship
        entity_relation.delete()

        # Recreate the relationship
        entity_relation = G(
            EntityRelationship,
            sub_entity=sub_entity,
            super_entity=super_entity
        )

        # Delete the relationship
        entity_relation.delete()

        # Assert that we only have the correct events
        events = EntityRelationshipActivationEvent.objects.order_by('time')
        self.assertEqual(len(events), 4)
        self.assertTrue(events[0].was_activated)
        self.assertFalse(events[1].was_activated)
        self.assertTrue(events[2].was_activated)
        self.assertFalse(events[3].was_activated)



from django.db import connection
from os.path import dirname


class EntityHistoryForeignKeys(object):
    def get_sql(self, name):
        return open(
            dirname(__file__) + '/' + name
        ).read()

    def enable_cascade_delete(self):
        with connection.cursor() as cursor:
            cursor.execute(self.get_sql('foreign_keys_cascade_delete.sql'))



from datetime import datetime
from django.db import models
from entity.models import Entity, EntityQuerySet, AllEntityManager


class EntityActivationEvent(models.Model):
    """
    Models an event of an entity being activated or deactivated.
    """
    entity = models.ForeignKey(
        Entity,
        help_text='The entity that was activated / deactivated',
        on_delete=models.CASCADE
    )
    time = models.DateTimeField(
        db_index=True,
        default=datetime.utcnow,
        help_text='The time of the activation / deactivation'
    )
    creation_time = models.DateTimeField(
        db_index=True,
        default=datetime.utcnow,
        help_text='The time that this entry was created')
    updated_time = models.DateTimeField(
        db_index=True,
        default=datetime.utcnow,
        help_text='The time that this entry was updated'
    )
    was_activated = models.BooleanField(
        default=None,
        help_text='True if the entity was activated, false otherwise'
    )

    class Meta:
        app_label = 'entity_history'


class EntityRelationshipActivationEvent(models.Model):
    """
    Models an event of an entity relationship being activated or deactivated. Technically, entity relationships
    are either created or deleted, however, we use the terms activated and deactivated for consistency.
    """
    sub_entity = models.ForeignKey(
        Entity,
        related_name='+',
        help_text='The sub entity in the relationship that was activated / deactivated',
        on_delete=models.CASCADE

    )
    super_entity = models.ForeignKey(
        Entity,
        related_name='+',
        help_text='The super entity in the relationship that was activated / deactivated',
        on_delete=models.CASCADE
    )
    time = models.DateTimeField(
        db_index=True,
        default=datetime.utcnow,
        help_text='The time of the activation / deactivation'
    )
    creation_time = models.DateTimeField(
        db_index=True,
        default=datetime.utcnow,
        help_text='The time that this entry was created'
    )
    updated_time = models.DateTimeField(
        db_index=True,
        default=datetime.utcnow,
        help_text='The time that this entry was updated'
    )
    was_activated = models.BooleanField(
        default=None,
        help_text='True if the entity was activated, false otherwise'
    )

    class Meta:
        app_label = 'entity_history'


def get_sub_entities_at_times(super_entity_ids, times, filter_by_entity_ids=None):
    """
    Constructs the sub entities of super entities at points in time.

    :param super_entity_ids: An iterable of super entity ids
    :param times: An iterable of datetime objects
    :param filter_by_entity_ids: An iterable of entity ids over which to filter the results
    :returns: A dictionary keyed on (super_entity_id, time) tuples. Each key has a set of all entity ids that were sub
       entities of the super entity during that time.
    """
    er_events = EntityRelationshipActivationEvent.objects.filter(super_entity_id__in=super_entity_ids).order_by('time')
    if filter_by_entity_ids:
        er_events = er_events.filter(sub_entity_id__in=filter_by_entity_ids)

    ers = {
        (se_id, t): set()
        for se_id in super_entity_ids
        for t in times
    }

    for t in times:
        # Traverse the entity relationship events in ascending time, keeping track of if a sub entity was in a
        # relationship before time t
        for er_event in [er for er in er_events if er.time < t]:
            if er_event.was_activated:
                ers[(er_event.super_entity_id, t)].add(er_event.sub_entity_id)
            else:
                ers[(er_event.super_entity_id, t)].discard(er_event.sub_entity_id)

    return ers


def get_entities_at_times(times, filter_by_entity_ids=None):
    """
    Constructs the entities that were active at points in time.

    :param times: An iterable of datetime objects
    :param filter_by_entity_ids: An iterable of entity ids over which to filter the results
    :returns: A dictionary keyed on time values. Each key has a set of all entity ids that were active at the time.
    """
    e_events = EntityActivationEvent.objects.order_by('time')
    if filter_by_entity_ids:
        e_events = e_events.filter(entity_id__in=filter_by_entity_ids)

    es = {
        t: set()
        for t in times
    }

    for t in times:
        # Traverse the entity events in ascending time, keeping track of if an entity was active before time t
        for e_event in [e for e in e_events if e.time < t]:
            if e_event.was_activated:
                es[t].add(e_event.entity_id)
            else:
                es[t].discard(e_event.entity_id)

    return es


class EntityHistoryQuerySet(EntityQuerySet):
    """
    A queryset that wraps around the get_sub_entities_at_times and get_entities_at_times functions.
    """
    def get_sub_entities_at_times(self, super_entity_ids, times):
        return get_sub_entities_at_times(
            super_entity_ids, times, filter_by_entity_ids=self.values_list('id', flat=True))

    def get_entities_at_times(self, times):
        return get_entities_at_times(times, filter_by_entity_ids=self.values_list('id', flat=True))


class AllEntityHistoryManager(AllEntityManager):
    def get_queryset(self):
        return EntityHistoryQuerySet(self.model)

    def get_sub_entities_at_times(self, super_entity_ids, times):
        return self.get_queryset().get_sub_entities_at_times(super_entity_ids, times)

    def get_entities_at_times(self, times):
        return self.get_queryset().get_entities_at_times(times)


class ActiveEntityHistoryManager(AllEntityHistoryManager):
    """
    The default 'objects' on the EntityHistory model. This manager restricts all Entity queries to happen over active
    entities.
    """
    def get_queryset(self):
        return EntityHistoryQuerySet(self.model).active()


class EntityHistory(Entity):
    """
    A proxy model for entities that overrides the default model manager. This model manager provides additional
    functionality to query entities and entity relationships at points in time.
    """
    class Meta:
        proxy = True

    objects = ActiveEntityHistoryManager()
    all_objects = AllEntityHistoryManager()



# -*- coding: utf-8 -*-

from django.db import models, migrations
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('entity', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='EntityActivationEvent',
            fields=[
                ('id', models.AutoField(serialize=False, auto_created=True, verbose_name='ID', primary_key=True)),
                ('time', models.DateTimeField(db_index=True, help_text='The time of the activation / deactivation')),
                ('was_activated', models.BooleanField(help_text='True if the entity was activated, false otherwise', default=None)),
                ('entity', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, help_text='The entity that was activated / deactivated', to='entity.Entity')),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='EntityRelationshipActivationEvent',
            fields=[
                ('id', models.AutoField(serialize=False, auto_created=True, verbose_name='ID', primary_key=True)),
                ('time', models.DateTimeField(db_index=True, help_text='The time of the activation / deactivation')),
                ('was_activated', models.BooleanField(help_text='True if the entity was activated, false otherwise', default=None)),
                ('sub_entity', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='entity.Entity', related_name='+', help_text='The sub entity in the relationship that was activated / deactivated')),
                ('super_entity', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='entity.Entity', related_name='+', help_text='The super entity in the relationship that was activated / deactivated')),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='EntityHistory',
            fields=[
            ],
            options={
                'proxy': True,
            },
            bases=('entity.entity',),
        ),
    ]



from datetime import datetime

from django.db import connection
from django.test import TransactionTestCase
from django_dynamic_fixture import G
from entity.models import Entity
from entity_history.models import EntityRelationshipActivationEvent, EntityActivationEvent


class CascadeDeleteTest(TransactionTestCase):
    def test_cascade_delete(self):
        # Create some entities and events
        super_e1 = G(Entity)
        super_e2 = G(Entity)
        sub_e1 = G(Entity)
        sub_e2 = G(Entity)

        G(EntityActivationEvent, was_activated=True, entity=super_e1, time=datetime(2013, 2, 1))
        G(EntityActivationEvent, was_activated=True, entity=super_e2, time=datetime(2013, 2, 1))
        G(EntityActivationEvent, was_activated=True, entity=sub_e1, time=datetime(2013, 2, 1))
        G(EntityActivationEvent, was_activated=True, entity=sub_e2, time=datetime(2013, 2, 1))

        G(
            EntityRelationshipActivationEvent, was_activated=True, super_entity=super_e1, sub_entity=sub_e1,
            time=datetime(2013, 2, 1))
        G(
            EntityRelationshipActivationEvent, was_activated=False, super_entity=super_e1, sub_entity=sub_e1,
            time=datetime(2013, 2, 3))
        G(
            EntityRelationshipActivationEvent, was_activated=False, super_entity=super_e1, sub_entity=sub_e1,
            time=datetime(2013, 2, 4))
        G(
            EntityRelationshipActivationEvent, was_activated=True, super_entity=super_e1, sub_entity=sub_e1,
            time=datetime(2013, 2, 4, 12))
        G(
            EntityRelationshipActivationEvent, was_activated=False, super_entity=super_e1, sub_entity=sub_e1,
            time=datetime(2013, 3, 4, 12))
        G(
            EntityRelationshipActivationEvent, was_activated=True, super_entity=super_e2, sub_entity=sub_e1,
            time=datetime(2013, 1, 1))
        G(
            EntityRelationshipActivationEvent, was_activated=False, super_entity=super_e2, sub_entity=sub_e1,
            time=datetime(2013, 12, 1))

        G(
            EntityRelationshipActivationEvent, was_activated=True, super_entity=super_e1, sub_entity=sub_e2,
            time=datetime(2013, 2, 4))
        G(
            EntityRelationshipActivationEvent, was_activated=False, super_entity=super_e1, sub_entity=sub_e2,
            time=datetime(2013, 2, 20))
        G(
            EntityRelationshipActivationEvent, was_activated=False, super_entity=super_e1, sub_entity=sub_e2,
            time=datetime(2013, 3, 4))
        G(
            EntityRelationshipActivationEvent, was_activated=True, super_entity=super_e1, sub_entity=sub_e2,
            time=datetime(2013, 3, 4, 12))
        G(
            EntityRelationshipActivationEvent, was_activated=True, super_entity=super_e1, sub_entity=sub_e2,
            time=datetime(2013, 3, 4, 13))
        G(
            EntityRelationshipActivationEvent, was_activated=True, super_entity=super_e2, sub_entity=sub_e2,
            time=datetime(2013, 1, 1))
        G(
            EntityRelationshipActivationEvent, was_activated=False, super_entity=super_e2, sub_entity=sub_e2,
            time=datetime(2013, 12, 1))

        # Run the delete outside of django
        with connection.cursor() as cursor:
            cursor.execute('DELETE FROM entity_entity')

        # Check that all tables are empty
        self.assertEqual(Entity.objects.all().count(), 0)
        self.assertEqual(EntityRelationshipActivationEvent.objects.all().count(), 0)
        self.assertEqual(EntityActivationEvent.objects.all().count(), 0)
